// Copyright 2014-2015, SAS Institute Inc., Cary, NC, USA.  All Rights Reserved.

package com.sas.deployment

import java.util.regex.Matcher
import groovy.transform.InheritConstructors

import com.sas.gate.CodeAccessor
import com.sas.gate.api.PipeCode
import com.sas.gate.api.template.*

/**
   <overview>Deployment code for SAS Visual Scenario Designer (scdsvrc)</overview>
   <p>
      This code configures the deployment of scdsvrc using what corresponds to
      the following Ant targets:
      <ul class='nodot'>
         <li>configure</li>
         <li>unconfigure</li>
      </ul>
   </p>

   @since      14w19
   @version    15w04
   @author     Prashant Ghorpade
   @author     Steve Amerige
   @see        #SOURCE#
*/

@InheritConstructors
class SCDSVRC extends CodeAccessor
{
   // ===================================================== GLOBAL

   /**
      Identifier

      @param id <type>String</type>: The scdsvrc name
   */
   String id = 'SCDSVRC'

   /**
      Callback for hook: <hook>properties &gt; init</hook>

      @param t__properties_get   <type>ArrayList&lt;String&gt;</type>: get object properties on each dependson property

   <ul class='nodot'>
      <li>levconfig</li>
      <li>workspacesrv</li>
   </ul>
   */
   def properties_init =
   {
      pass.with {

         debug = ['debug']

         t__properties_get =
         [
            'scdsvrc.dependson.levconfig.fqid',
            'scdsvrc.dependson.dsbatchsrv.fqid',
            'scdsvrc.dependson.bihpgrdc.fqid',
            'scdsvrc.dependson.webinfdsvrc.fqid'
         ]

         t__retrievewipdbcredentials =
         [
               enabled: true
         ]
      }
   }

   /**
      Callback for hook: <hook>properties &gt; post</hook>

      @param filters <type>Map&lt;String&gt;</type>: tokens for string substitution
      @param t__setobjectproperties    <type>ArrayList&lt;String&gt;</type>: Properties that will be stored in the configured objects text store
      @param t__createdirectories <type>Map&lt;String&gt;</type>: directories to be created in configure target<br>
            <indent><key>prefix</key>- <type>String</type>: base directory to create<br>
            <key>entries</key>- <type>ArrayList&lt;String&gt;</type>: directories and subdirectories to be created<br></indent>
      <pre>
   /install/config/Lev1/Applications/<your_product_name_here>/<your_version>
|-- custom
|   |-- config
|   `-- source
|-- data
|-- install
|-- lasr
|-- sasmisc
`-- scenario
    `-- scenario_source</pre>

      @param t__copyfiles         <type>ArrayList&lt;Map&gt;</type>: files to copy and conditions for the copy<br>
            <indent><key>src</key>- <type>String</type>: Source directory to copy from<br>
            <key>include</key>- <type>String</type>: Regex pattern of file(s) to copy<br>
            <key>dst</key>- <type>String</type>: Destination directory for files to be copied to<br>
            <key>strip</key>- <type>String</type> <kind>optional</kind>: Remove extension form file name i.e., remove .orig extension<br>
            <key>tokens</key>- <type>String</type> <kind>optional</kind>: filterset to use for token replacement<br></indent>
      @param t__loadmetadata_createfolders   -
      @param t__importpackages_createfolders   -
      @param t__loadmetadata_load   <type>String</type>: Fully qualified path to the token substituted appxml file
      @param t__sasjobs_run      -
   */
   def properties_post =
   {
      // Simple Strings
        String releaseNumber                      = api.'scdsvrc.release.number'
        String legalName                          = api.'scdsvr.legal.name'
        String normalizedOfferingLegalName        = api.'scdsvrc.normalized.offering.name'

      // Config Paths
        String configCustCfgDir                   = canonical("$configDir/custom/config")

      // Install Paths (Dirs and Files)
        String installSASDir                      = api.'sasinstall.install.dir'

      // Other
      String classIdentifier                    = api.'scdsvrc.configobj.classid'
      String majorNumber                        = api.'scdsvrc.major.number'
      String minorNumber                        = api.'scdsvrc.minor.number'
      String cmacrosscdsvr                      = api.'scdsvrc.install.cmacros.dir'
      String configlevappdir                    = api.'config.lev.appservercontext.dir'

      String fullLocale                         = api.'sashome.locale.name'
      String shortLocale                        = fullLocale.substring(0, fullLocale.indexOf('_'))
      String stageLocale                        = (shortLocale == 'zh') ? '_' + fullLocale : '_' + shortLocale
      String locale                             = (stageLocale == '_en') ? '' : stageLocale

      pass.with
      {
         _sasConfigDir                       = api.'config.lev.applications.dir'
         _configDir                          = canonical("$_sasConfigDir/$normalizedOfferingLegalName")
         _batchServerDir                     = canonical("$configlevappdir/BatchServer")
         _applicationFolder                  = api.'scdsvrc.application.folder'
         _installSasmiscDir                  = api.'scdsvrc.install.sasmisc.dir'
         _installSasmiscxmlDir               = canonical("$_installSasmiscDir/xml")
         _installSasmiscDDLPath              = canonical("$_installSasmiscDir/ddl/postgresql/vsd_ddl.sql.orig")
         _installConfigDir                   = api.'scdsvrc.install.config.dir'
         _appxmlFile                         = canonical("$_configDir/install/${pass.'12byte'}.appxml")
         _applicationSubFolder               = api.'scdsvrc.application.sub.folder'
         _softwareComponentName              = api.'scdsvrc.software.component.name'
         _sampleCheckBox                     = getProperty('scdsvrc.dbms.samples', '0') { toBoolean(it) }
         _locale                             = "$locale"
         _offeringLegalName                  = api.'scdsvrc.offering.legal.name'
         _configInstallDir                   = canonical("$_configDir/install")
         _autoexecPath                       = canonical("$configCustCfgDir/autoexec.sas")
         _configucmacrosDir                  = canonical("$_configDir/ucmacros")
         _scdsvrcdbmsname                    = api.'scdsvrc.dbms.name'
         _dbmsUserid                         = api.'scdsvrc.dbms.userid'
         _ucmacrosscdsvr                     = api.'scdsvrc.install.ucmacros.dir'

         // Properties
         api.createProperty('scdsvrc.autoexec.path', _autoexecPath)
         api.createProperty('scdsvrc.config.dir', _configDir)
         api.createProperty('scdsvrc.configobj.name',_softwareComponentName)
         api.createProperty('scdsvrc.batchserver.file', canonical("$pass._batchServerDir/autoexec_usermods.sas"))

         _filters =
         [
            'scdsvr.app.home'                      : _configDir,
            'scdsvr.software.component.name'       : _softwareComponentName,
            'class.identifier'                     : classIdentifier,
            'description'                          : _softwareComponentName,
            'dbms.userid'                          : _dbmsUserid,
            'legal.name'                           : legalName,
            'major.number'                         : majorNumber,
            'minor.number'                         : minorNumber,
            'release.number'                       : releaseNumber,
            'cmacros.scdsvr'                       : cmacrosscdsvr,
            'ucmacros.scdsvr'                      : _ucmacrosscdsvr,
            'sasmisc.scdsvr'                       : _installSasmiscDir,
            'software.component.name'              : _softwareComponentName,
            'product.version.folder'               : _applicationSubFolder,
            'scdsvr.home'                          : installSASDir,
            'normalized.legal.name'                : normalizedOfferingLegalName,
            'sasbatchserver.dir'                   : _batchServerDir,
         ]
      }
  }

  def configure__properties_post =
   {
     pass.with {

         t__createdirectories =
         [
            prefix: _configDir,
            entries:
            [
                 'install',
                 'logs',
                 'deployments',
                 'ucmacros'
            ]
         ]

         t__setobjectproperties =
         [
           properties:
           [
            'scdsvrc.configobj.classid',
            'scdsvrc.configobj.version',
            'scdsvrc.configobj.name',
            'scdsvrc.release.number',
            'scdsvrc.dbms.userid',
            'scdsvrc.offering.legal.name',
            'scdsvrc.software.component.name',
            'scdsvrc.dbms.data.connection.driver_class',
            'scdsvrc.dbms.name',
            'scdsvrc.dbms.samples',
           ]
        ]

         t__copyfiles =
         [
            [  src:        _installConfigDir,
               include:    ~/vsd\.sas/,
               dst:        _batchServerDir,
               tokens:     _filters
            ],
            [  src:        _installConfigDir,
               include:    ~/autoexec_usermods\.sas\.orig/,
               dst:        api.'temp.dir',
               strip:      '.orig',
               tokens:     _filters
            ],
            [  src:        _ucmacrosscdsvr,
               include:    ~/vsd_set_rundt\.sas/,
               dst:        _configucmacrosDir,
               tokens:     _filters
            ],
            [  src:        "$_installSasmiscDir/ddl/postgresql",
               include:    ~/.*\.orig/,
               dst:        _configInstallDir,
               strip:      '.orig',
               tokens:     _filters
            ],
            [  src:        "$_installConfigDir/Deployment/Metadata",
               include:    ~/scdsvrc\.appxml\.orig/,
               dst:        _configInstallDir,
               strip:      '.orig',
               tokens:     _filters
            ],
            [  src:        "$_installSasmiscDir/xml",
               include:    ~/.*\.orig/,
               dst:        _installSasmiscxmlDir,
               strip:      '.orig',
               tokens:     _filters
            ],
         ]

         t__loadmetadata_createfolders =
         [
            _applicationFolder
         ]

         t__loadmetadata_load = _appxmlFile

         t__setmetadataperms =
         [
            perm  : 'G',
            group : 'SASUSERS',
            set:
            [
               [
                  parentfolder: _offeringLegalName,
                  folders: _softwareComponentName,
                  permNames:
                  [
                     'ReadMetadata','WriteMetadata','WriteMemberMetadata'
                  ]
               ],
            ]
         ]
      }
   }

  def configure_database_init__properties_post =
   {
     pass.with {
         t__database_createwipdatabase =
         [
            wip: api.'scdsvrc.dbms.name'
         ]

         t__database_createwipdbuser =
         [
            wip         : api.'scdsvrc.dbms.name',
            userid      : api.'scdsvrc.dbms.userid',
            passwd      : api.'scdsvrc.dbms.passwd',
         ]
      }
   }

   def migration__properties_post =
   {
      pass.with {

               t__database_restorewipdatabase =
               [
                  executable: canonical("${api.'webinfdsvrc.install.dir.bin'}/psql"),
                  resultproperty: '_!returnCode',
                  input: _migrationSQLPath,
                  failonerror: true,
                  env:
                  [
                        PGPASSWORD:          api.'scdsvrc.webinfdsvrc.admin.passwd',
                        'LD_LIBRARY_PATH':   canonical("${api.'webinfdsvr.install.dir'}/lib"),
                        'LIBPATH':           canonical("${api.'webinfdsvr.install.dir'}/lib"),
                  ],
                  arg:
                  [
                     "--dbname=${api.'scdsvrc.dbms.name'}",
                     "--username=${api.'iomsrv.webinfdsvrc.userid'}",
                     "--host=${api.'iomsrv.webinfdsvrc.host'}",
                     "--port=${api.'iomsrv.webinfdsvrc.port'}"
                  ]

               ]

               t__database_grantwipuser =
               [
                adminUserId:  api.'scdsvrc.webinfdsvrc.admin.userid',
                adminPasswd:  api.'scdsvrc.webinfdsvrc.admin.passwd',
                dbHost:       api.'scdsvrc.webinfdsvrc.admin.host',
                dbPort:       api.'scdsvrc.webinfdsvrc.admin.port',
                grants:
                  [
                     [ // NOTE: Do not change this element it is used by code in configure__database_init
                     name    : _dbmsUserid,
                     dbname  : _scdsvrcdbmsname
                     ],
                     [
                     name    : _dbmsUserid,
                     dbname  : _scdsvrcdbmsname,
                     grant   : "all privileges on database $_scdsvrcdbmsname" as String,
                     ]
                  ]
               ]

               t__database_updatewipdatabase =
               [
                databases:
                   [
                     [
                      enabled     : _sampleCheckBox,
                      wip         : api.'scdsvrc.dbms.name',
                      userid      : api.'scdsvrc.dbms.userid',
                      passwd      : api.'scdsvrc.dbms.passwd',
                      file        : "$_installSasmiscDir/ddl/postgresql/vsd_load_sample" + '${locale}.sql'
                     ]
                   ]
               ]
         }
   }

   def database_patch__properties_post =
   {
      pass.with {
       if (_fromVersion == '6.2')
        {
          t__database_updatewipdatabase =
          [
            databases:
            [
               [
                  wip         : api.'scdsvrc.dbms.name',
                  userid      : api.'scdsvrc.dbms.userid',
                  passwd      : api.'scdsvrc.dbms.passwd',
                  file        : canonical("$_configInstallDir/vsd_patch-6.2-6.3.sql")
                ]
             ]
          ]
       }

      if (_migratedFromVersion == '6.2')
        {
          t__database_updatewipdatabase =
          [
            databases:
            [
               [
                  wip         : api.'scdsvrc.dbms.name',
                  userid      : api.'scdsvrc.webinfdsvrc.admin.userid',
                  passwd      : api.'scdsvrc.webinfdsvrc.admin.passwd',
                  file        : canonical("$_configInstallDir/vsd_patch-6.2-6.3.sql")
                ]
             ]
          ]
       }
     }
   }

   def configure_database__properties_post =
   {
      pass.with {
               t__database_updatewipdatabase =
               [
                databases:
                   [
                    [
                     wip         : api.'scdsvrc.dbms.name',
                     userid      : api.'scdsvrc.dbms.userid',
                     passwd      : api.'scdsvrc.dbms.passwd',
                     file        : canonical("$_configInstallDir/vsd_ddl.sql")
                     ],
                     [
                      enabled     : _sampleCheckBox,
                      wip         : api.'scdsvrc.dbms.name',
                      userid      : api.'scdsvrc.dbms.userid',
                      passwd      : api.'scdsvrc.dbms.passwd',
                      file        : "$_installSasmiscDir/ddl/postgresql/vsd_load_sample" + '${locale}.sql'
                     ]
                   ]
                ]
           }
    }

   // ===================================================== TARGET: configure

   /**
      Target: configure

      @see com.sas.gate.api.template.ConfigureTemplate
   */
   def configure()
   {
    //api.stop()

      ConfigureTemplate t = new ConfigureTemplate(this)

      t['migration > init']                  .add(configure__migration_init)
      t['properties > init']                 .add(properties_init)
      t['properties > post']                 .add(properties_post)
      t['properties > post']                 .add(configure__properties_post)
      t['properties > post']                 .add(configure_database_init__properties_post)
      t['properties > post']                 .add(configure_database__properties_post).when('!t__migrate')
      t['properties > post']                 .add(migration__properties_post).when('t__migrate')
      t['properties > post']                 .add(configure__setobjectproperties_init).when('t__migrate')
      t['properties > post']                 .add(database_patch__properties_post).when('t__migrate')
      t['loadmetadata > migration > init']   .add(configure__loadmetadata_migration_init).when('t__migrate')
      t['copyfiles']                         .add(callback:configure__copyfiles, after:t.t__copyfiles)
      t['database > init']                   .add(configure__database_init).when('t__migrate')
      t['docparts > init']                   .add(configure__docparts_init)
      t['docparts > migration']              .add(configure__docparts_migration).when('t__migrate')

      t.execute()

     //api.stop()
   }

   def configure__migration_init =
   {
     pass.with
      {
         t__migrate = getProperty('scdsvrc.migration.is_enabled', '0') { toBoolean(it) }

         if (t__migrate)
         {
            _migratedFromVersion  = api.'scdsvrc.migrated.version'

            t__searchandcall =
            [
               fqids:
               [
                  [
                     actions: [ __template.t__deletemetadata ],
                     type:    'Tree',
                     search:  "@Name = 'Visual Scenario Design Svr 6.2'"
                  ]
                ]
             ]
            if (_migratedFromVersion == '6.2')
               _migrationSQLPath = canonical("${api.'config.migration.package.tier.dir'}/scdsvrc/vsd62_migration.sql")
            else if (_migratedFromVersion == '6.3')
               _migrationSQLPath = canonical("${api.'config.migration.package.tier.dir'}/scdsvrc/vsd63_migration.sql")
         }
      }
   }

   def configure__loadmetadata_migration_init =
   {
     pass.with {
         if (_migratedFromVersion == '6.2')
         {
            t__loadmetadata_migration_createfolders =
            [
               _applicationSubFolder
            ]
         }

         t__loadmetadata_migration_update =
         [
               file:             _appxmlFile,
               folder:           _applicationSubFolder,
               applicationname:  _softwareComponentName
         ]

      }
   }

   def configure__copyfiles =
   {
      String baseFilename = canonical("${pass._batchServerDir}/autoexec_usermods.sas")
      String saveAs = "${baseFilename}.orig"
      File saveAsFile = new File(saveAs)

      if (!(saveAsFile.exists()))
             new File(baseFilename).renameTo(saveAsFile)

      def concatItems =
      [
         src:
         [
            [
               dir: pass._batchServerDir,
               include: ~/autoexec_usermods\.sas\.orig/
            ],
            [
               dir: api.'temp.dir',
               include: ~/autoexec_usermods\.sas/
            ]
         ],
         overwrite:true,
         dst: baseFilename
      ]
      new PipeCode(concatItems).concat().write()
   }

   def configure__database_init =
   {
     pass.with {
         File f = new File(_installSasmiscDDLPath)
         def regex =
         [
            /grant (all privileges on table [^ ]*) to .*/,
            /grant (usage on all sequences .*) to .*/
         ]

         ArrayList<String> grants = []
         f.eachLine { line ->
            regex.each { item ->
                Matcher m = line =~ item
                if (m.matches())
                    grants += m.group(1) as String
            }
         }

         t__database_grantwipuser.grants +=
            t__database_grantwipuser.grants[0] + [grant:grants]
      }
   }

   /**
      Callback for hook: <hook>properties &gt; post</hook>

      @param t__docparts   <type>ArrayList&lt;Map&gt;</type>: tokens in Instructions.html to replace and properties to replace them with
            <indent><key>replace, prepend, append</key>- <type>String</type>: Type of replacement to perform<br>
            <indent><key>token</key>- <type>String</type>: Token in Instructions.html to replace<br>
            <key>value</key>- <type>String</type>: Property to replace token with</indent></indent>

   */
   def configure__docparts_init =
   {
     pass.with
      {
         t__docparts =
         [
         actions:
            [
               [replace:[
                  token:'instructions.scdsvrc.head',
                  value:api.'instructions_scdsvrc_head.txt']],
               [replace:[
                  token:'instructions.scdsvrc.details.table.head',
                  value:api.'instructions_scdsvrc_details_table_head.txt']],
               [prepend:[
                  token:'instructions.scdsvrc.details.table.item',
                  value:api.'instructions_scdsvrc_preparation.txt']],
               [prepend:[
                  token:'instructions.scdsvrc.details.table.item',
                  value:api.'instructions_scdsvrc_file_locations.txt']],
                [prepend:[
                  token:'instructions.scdsvrc.details.table.item',
                  value:api.'instructions_scdsvrc_details_logging.txt']],
               [replace:[
                  token:'instructions.scdsvrc.details.table.tail',
                  value:api.'instructions_scdsvrc_details_table_tail.txt']],
            ]
         ]
      }
   }

   def configure__docparts_migration =
   {
     pass.with {
         t__docparts.actions +=
         [replace:[
            token:'instructions.scdsvrc.migration.report',
            value: api.'scdsvrc_migrateconfig_header_intro.txt']
         ]
      }
   }

   // ===================================================== TARGET: updateConfigure

   /**
      Target: updateConfigure

      @see com.sas.gate.api.template.ConfigureTemplate
   */
   def updateConfigure()
   {
    //api.stop()

      pass.with {
      _maintenanceFrom = getProperty("${pass.'12byte'}.maintenance.from.version", '')
      _fromVersion      = api.baseVersion(_maintenanceFrom)
      _maintenanceTo    = getProperty("${pass.'12byte'}.maintenance.to.version", '')
      _toVersion        = api.baseVersion(_maintenanceTo)

      api.createProperty('scdsvrc.maintenance.from.release', "${api.baseVersion(_maintenanceFrom)}")

      if (_maintenanceFrom)
      {
         if (_maintenanceFrom != _maintenanceTo)
         {
            if (_fromVersion != _toVersion)
            {
               UpdateConfigureTemplate t = new UpdateConfigureTemplate(this)

               t['properties > pre']               .add(updateconfigure__pre_init)
               t['properties > init']              .add(properties_init)
               t['properties > post']              .add(properties_post)
               t['properties > post']              .add(configure__properties_post)
               t['properties > post']              .add(database_patch__properties_post)
               t['properties > post']              .add(updateconfigure__properties_post)
               t['loadmetadata > init']            .add(updateconfigure__loadmetadata_init)
               t['setobjectproperties > init']     .add(configure__setobjectproperties_init)
               t['copyfiles']                      .add(callback:configure__copyfiles, after:t.t__copyfiles)
               t['move > init']                    .add(updateconfigure__movedirsondisk_init)
               t['docparts > init']                .add(updateconfigure__docparts_init)

               t.execute()
            }
            else
            {
               UpdateConfigureTemplate t = new UpdateConfigureTemplate(this)

               t['properties > pre']               .add(updateconfigure__pre_init)
               t['properties > init']              .add(properties_init)
               t['properties > post']              .add(properties_post)
               t['properties > post']              .add(configure__properties_post)
               t['properties > post']              .add(updateconfigure__properties_post)
               t['loadmetadata > init']            .add(updateconfigure__loadmetadata_init)
               t['setobjectproperties > init']     .add(configure__setobjectproperties_init)
               t['copyfiles']                      .add(callback:configure__copyfiles, after:t.t__copyfiles)
               t['move > init']                    .add(updateconfigure__movedirsondisk_init)
               t['docparts > init']                .add(updateconfigure__docparts_init)

              //Add i10n doc-parts here

               t.execute()
            }
         }
      }
      else
         configure()
    }
   //api.stop()
  }


  def updateconfigure__pre_init =
   {
       String legalName = api.createProperty('scdsvrc.short.legal.name', 'Visual Scenario Design Svr')

       String version = api.baseVersion(api.'scdsvrc.maintenance.to.version', api.'scdsvrc.release.number')
       api.createProperty('scdsvrc.release.number', version)
       api.createProperty('scdsvrc.configobj.version', version)

       String softwareComponent = api.createProperty('scdsvrc.software.component.name', "$legalName $version")
       String applicationFolder = api.'scdsvrc.application.folder'
       String oldFolderName = "$applicationFolder/$legalName 6.2"
       api.createProperty('scdsvrc.release.number', version)
       String normalizedOfferingLegalName       = api.'scdsvrc.normalized.offering.name'
       String sasConfigDir                      = api.'config.lev.applications.dir'
       String configDir                         = canonical("$sasConfigDir/$normalizedOfferingLegalName")
       api.createProperty('scdsvrc.config.dir', configDir)

       String subFolder = api.createProperty('scdsvrc.application.sub.folder', "$applicationFolder/$softwareComponent")

       pass.with
        {
         if ('$_fromVersion == 6.2')
          {
             t__searchandcall =
             [
               fqids:
               [
                  [
                    actions: [ __template.t__deletemetadata ],
                    type:    'Tree',
                    search:  "@Name = 'Visual Scenario Design Svr 6.2'"
                  ]
               ]
             ]
          }
      }
  }

  def updateconfigure__loadmetadata_init =
   {
    pass.with {
        t__loadmetadata_update =
        [
             file:             _appxmlFile,
             folder:           _applicationSubFolder,
             applicationname:  _softwareComponentName
        ]
      }
    }

  def configure__setobjectproperties_init =
    {
      pass.with {
        t__setobjectproperties +=
        [
           existingproperties: 'replace'
        ]
      }
    }

  def updateconfigure__properties_post =
   {
    pass.with {
       t__setconfigureversion =
       [
           fqid:     api.'scdsvrc.configobj.fqid',
           version:  _maintenanceTo
       ]
     }
   }

  def updateconfigure__movedirsondisk_init =
  {
     pass.with {
      if (_fromVersion == '6.2' )
      {
        t__movedirsondisk =
        [
          move:
          [
            src:        canonical("$_sasConfigDir/SASVisualScenarioDesigner6.2"),
            dst:        _configDir,
            newName:    '6.2'
          ]
        ]
       }
     }
  }

/**
      Callback for hook: <hook>properties &gt; post</hook>

      @param t__docparts   <type>ArrayList&lt;Map&gt;</type>: tokens in Instructions.html to replace and properties to replace them with
            <indent><key>replace, prepend, append</key>- <type>String</type>: Type of replacement to perform<br>
            <indent><key>token</key>- <type>String</type>: Token in Instructions.html to replace<br>
            <key>value</key>- <type>String</type>: Property to replace token with</indent></indent>

   */
   def updateconfigure__docparts_init =
   {
     pass.with
      {
         String text = (_fromVersion == '6.2' ? api.'updateconfig_scdsvrc_details_table_item_note_62.txt' : api.'updateconfig_scdsvrc_details_table_item_note_63.txt')

         t__docparts =
         [
            actions:
            [
               [replace:[
                  token:'instructions.scdsvrc.head',
                  value:api.'instructions_scdsvrc_head.txt']],
               [prepend:[
                  token:'instructions.scdsvrc.details.table.head',
                  value:api.'instructions_scdsvrc_details_table_head.txt']],
               [prepend:[
                  token:'instructions.scdsvrc.details.table.item',
                  value:api.'updateconfig_scdsvrc_details_table_item_header1.txt']],
               [prepend:[
                  token:'instructions.scdsvrc.details.table.item',
                  value:api.'updateconfig_scdsvrc_details_table_item_header2.txt']],
               [prepend:[
                  token:'instructions.scdsvrc.details.table.item',
                  value: text]],
               [replace:[
                  token:'instructions.scdsvrc.details.table.tail',
                  value:api.'instructions_scdsvrc_details_table_tail.txt']],
            ]
         ]
      }
   }

   // ===================================================== TARGET: checkForErrors

   /**
      Target: checkForErrors

      @see com.sas.gate.api.template.CheckForErrorsTemplate
   */
   def checkForErrors()
   {
      CheckForErrorsTemplate t = new CheckForErrorsTemplate(this)
      t.execute()
   }

   // ===================================================== TARGET: reportFailures

   /**
      Target: reportFailures

      @see com.sas.gate.api.template.ReportFailuresTemplate
   */
   def reportFailures()
   {
      ReportFailuresTemplate t = new ReportFailuresTemplate(this)
      t.execute()
   }

   // ===================================================== TARGET: unconfigure

   /**
      Target: unconfigure

      @see com.sas.gate.api.template.UnconfigureTemplate
   */
   def unconfigure()
   {
      //api.stop()

      UnconfigureTemplate t = new UnconfigureTemplate(this)

      t['properties > init']  .add(properties_init)
      t['properties > post']  .add(properties_post)
      t['init']               .add(unconfigure__init)
      t['docparts > init']    .add(unconfigure__docparts_init)

      t.execute()

      //api.stop()
   }

   def unconfigure__init =
   {
      String vsdsasfile    = canonical("$pass._batchServerDir/vsd.sas")
      String origFile      = canonical("$pass._installConfigDir/autoexec_usermods.sas.orig")
      String sourceFile    = canonical("$pass._batchServerDir/autoexec_usermods.sas")

      pass.with
      {
         t__removemetadatabypath =
         [
            paths:
            [
               [
                  objectPath     : _applicationFolder,
                  deleteContents : true
               ],
            ]
         ]

         t__delete_files =
         [
           paths:
            [
               vsdsasfile
            ]
         ]

         t__database_stopwipdbconn =
         [
            executable: canonical("${api.'webinfdsvrc.install.dir.bin'}/psql"),
            resultproperty: '_!returnCode',
            input: "$_installSasmiscDir/ddl/postgresql/vsd_kill_conn.sql",
            failonerror: true,
            env:
            [
                  PGPASSWORD:          api.'scdsvrc.webinfdsvrc.admin.passwd',
                  'LD_LIBRARY_PATH':   canonical("${api.'webinfdsvr.install.dir'}/lib"),
                  'LIBPATH':           canonical("${api.'webinfdsvr.install.dir'}/lib"),
            ],
            arg:
            [
               "--dbname=postgres",
               "--username=${api.'iomsrv.webinfdsvrc.userid'}",
               "--host=${api.'iomsrv.webinfdsvrc.host'}",
               "--port=${api.'iomsrv.webinfdsvrc.port'}"
            ]
         ]

         t__deletewipdatabase =
         [
            wip: api.'scdsvrc.dbms.name'
         ]

         t__deletewipuser =
         [
            username: api.'scdsvrc.dbms.userid'
         ]

         t__revertfilecontents =
         [
            src: sourceFile,
            remove:
            [
               [file: origFile]
            ]
         ]
      }
   }

   /**
      Callback for hook: <hook>properties &gt; post</hook>

      @param t__docparts   <type>ArrayList&lt;Map&gt;</type>: tokens in ConfigurationRemoval.html to replace and properties to replace them with
            <indent><key>replace, prepend, append</key>- <type>String</type>: Type of replacement to perform<br>
            <indent><key>file</key>- <type>String</type>: File to replace tokens in.<br>
            <key>token</key>- <type>String</type>: Token in Instructions.html to replace<br>
            <key>value</key>- <type>String</type>: Property to replace token with</indent></indent>

   */
   def unconfigure__docparts_init =
   {
      pass.with
      {
         t__docparts =
         [
            actions:
            [
               [replace:[
                  file:    api.'config.lev.unconfiguration.instructions.file',
                  token:   'unconfiguration.summary.scdsvrc.head',
                  value:   api.'unconfiguration_summary_scdsvrc_head.txt']],
               [prepend:[
                  file:    api.'config.lev.unconfiguration.instructions.file',
                  token:   'unconfiguration.summary.scdsvrc.item',
                  value:   api.'unconfiguration_summary_scdsvrc_item.txt']],
               [replace:[
                  file:    api.'config.lev.unconfiguration.instructions.file',
                  token:   'unconfiguration.summary.scdsvrc.tail',
                  value:   api.'unconfiguration_summary_scdsvrc_tail.txt']],
            ]
         ]
      }
   }

// ===================================================== TARGET: changeHost

   def changeHost()
   {
    //api.stop(properties:true)
      ChangeHostTemplate t = new ChangeHostTemplate(this)

      t['properties > init']  .add(properties_init)
      t['properties > post']  .add(properties_post)
      t['changehost > init']  .add(changehost__init)

      t.execute()
    //api.stop(properties:true)
   }

   def changehost__init =
   {
      Properties props = new Properties()
      File propsFile = new File(api.'configuration.properties.file')
      props.load(propsFile.newDataInputStream())
      api.getObjectProperties(prefix:'_!orig.',fqid:props.'scdsvrc.configobj.fqid')
   }
}
